# -*- coding: utf-8 -*-

# DocumentWrapper.py
# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import gtk, gobject, re
from completion import CompletionProvider
from AutoCloseHelper import AutoCloseHelper
import const

LEVELS = ["part", "chapter", "section", "subsection", "subsubsection", "paragraph", "subparagraph"]

# keys recognized, when the completion popup is visible
validCompletionKeys = ["Up", "Down", "Page_Up", "Page_Down", "Escape", "Return"]

class DocumentWrapper(gobject.GObject):
	"""
	Encapsulates an open document
	"""
	__gsignals__ = {
			"environment-changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ()),
			"structure-updated" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
	}
	
	dviCompiled = False		# set from __init__ after compiling 
							# to determine whether to enable DVI search
	
	environment = const.UNKNOWN
	
	_uri = None
	_changed = True
	
	_masterFilename = None
	
	# these re's may be compiled on demand
	reIndent, reStructure, reLabelOrBibitem, reNewcommand, reEnvironment, reDocument, \
			reArgRef = None, None, None, None, None, None, None
	messageDialog = None
	
	def __init__(self, textView, instance):
		gobject.GObject.__init__(self)
		
		self.reComment = re.compile(r"%.*?$", re.M)
		self.rePackage = re.compile(r"\\usepackage(\[(.*?)\])?\{(.*?)\}")
		
		self._instance = instance
		self._textView = textView
		self.textBuffer = textView.get_buffer()
		
		# copy function into local namespace to avoid lookups
		self.gdk_keyval_name = gtk.gdk.keyval_name
		
		self.autoCloseEnabled = instance.preferences.setdefault("/auto-closing-enabled", True)
		
		self.modelStructure = gtk.TreeStore(str, str)					# name, complete string
		self.modelLabelsAndBibs = gtk.ListStore(str, int, str)			# name, type (0:label, 1:bibitem), complete string
		self.modelNewcommands = gtk.ListStore(str, str, str, str)		# name, definition, markup (name), complete string
		self.modelPackages = gtk.ListStore(str, str)					# package, options
		
		self.completionProvider = CompletionProvider(instance.builtinCompletionTargets, self.modelNewcommands, self.modelLabelsAndBibs, self.modelPackages)
		
		self.autoCloseHelper = AutoCloseHelper(self)
		
		self.textBuffer.connect("changed", self.__onBufferChanged)
		textView.connect("key-release-event", lambda *a: self.__recognizeEnvironment())
		textView.connect("button-release-event", lambda *a: self.__recognizeEnvironment())
		
		if instance.preferences.setdefault("/auto-completion-enabled", True):
			textView.connect_after("key-release-event", self.__onKeyReleaseEvent)
			textView.connect("key-press-event", self.__onKeyPressEvent)
			textView.connect("focus-out-event", self.__onFocusOutEvent)
			textView.connect("button-press-event", self.__onButtonPressed)
	
	def _isFragment(self):
		""" Returns True if this document seems to be only a fragment """
		string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter(), False)
		return (string.find(r"\begin{document}") == -1)
		
	def _getMasterFilename(self):
		""" Returns the filename of this document's master """
		# return cached filename if available
		if self._masterFilename:
			return self._masterFilename
		# reset
		self._masterFilename = None
		
		ownFilename = self.getFilename()
		if ownFilename:	
			import os, glob
			docName = os.path.splitext(os.path.basename(ownFilename))[0]
			for filename in glob.glob(os.path.dirname(ownFilename) + "/*.tex"):
				contents = open(filename).read()
				if contents.find(r"\include{" + docName) != -1:		# this one includes the file
					if contents.find(r"\begin{document}") != -1:	# and it's a full document >> bingo!
						self._masterFilename = filename
					break
				
		return self._masterFilename
	
	def __onButtonPressed(self, textView, event):
		self._instance.completionPopup.hide()
	
	def __onFocusOutEvent(self, textView, event):
		self._instance.completionPopup.hide()
	
	def __onKeyPressEvent(self, textView, event):
		keyVal = self.gdk_keyval_name(event.keyval)	
		autoCompletePopup = self._instance.completionPopup
		
		if autoCompletePopup.get_property("visible"):
			if keyVal == "Return":
				# accept the target
				target = autoCompletePopup.getSelectedTarget()
				self.textBuffer.insert_at_cursor(target[len(self.completionProvider.lastPrefix):])
				if self.autoCloseEnabled:
					self.autoCloseHelper.autoClose()
				autoCompletePopup.hide()
				return True
			elif keyVal == "Escape":
				# cancel autocompletion
				autoCompletePopup.hide()
				return True
			
			# move the cursor
			try:
				delta = {"Up" : -1, "Down" : 1, "Page_Up" : -5, "Page_Down" : 5}[keyVal]
				autoCompletePopup.moveCursor(delta)
				return True
			except KeyError:
				pass
	
	def __onKeyReleaseEvent(self, textView, event):
		""" Try to complete a written command using self.completionProvider and show 
			the self._instance.completionPopup """
		keyVal = self.gdk_keyval_name(event.keyval)
		autoCompletePopup = self._instance.completionPopup
		
		if keyVal == "Escape":
			## abort auto complete
			return True
		if autoCompletePopup.get_property("visible"):
			if keyVal in validCompletionKeys:
				## Signal bezieht sich aufs Popup, nicht weitergeben
				return True
		else:
			## do not auto complete after arrow keys have been pressed
			if keyVal in ["Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
				return False
		
		if keyVal == "braceright":
			autoCompletePopup.hide()
			#self.__autoClose()
			if self.autoCloseEnabled:
				self.autoCloseHelper.autoClose()
			return False
		
		insIter = self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert())
		
		if not insIter.ends_word():
			autoCompletePopup.hide()
			return False
		
		# look for command
		limitIter = insIter.copy()
		limitIter.backward_chars(30)
		match = insIter.backward_search("\\", False, limitIter)
		if not match:
			autoCompletePopup.hide()
			return False
		
		# FIXME: this is very bad, signal handlers should not be connected until 
		# there is LaTeX in the document to save resources
		if not self.isLatex():
			autoCompletePopup.hide()
			return False
		
		string = self.textBuffer.get_text(insIter, match[1], False)
		
		## find targets
		completionList = self.completionProvider.complete(string)
		nMatches = len(completionList)
		if not nMatches or string in [match[0] for match in completionList]:
			autoCompletePopup.hide()
			return False
		
		## store matches in the model
		autoCompletePopup.updateList(completionList)
		
		# calculate iter coords
		rect = self._textView.get_iter_location(self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert()))
		x, y = self._textView.buffer_to_window_coords(gtk.TEXT_WINDOW_WIDGET, rect.x, rect.y)
		
		# add widget coords
		rootX, rootY = self._textView.get_window(gtk.TEXT_WINDOW_WIDGET).get_geometry()[:2]
		x += rootX
		y += rootY
		
		# add window coords
		rootX, rootY = self._instance.window.get_position()
		x += rootX
		y += rootY
		
		# move and show the popup
		autoCompletePopup.move(x + 5, y + 42)
		autoCompletePopup.show_all()
		
		return True
	
	def activate(self):
		""" Called when the document is activated on tab change """
		uri = self.textBuffer.get_uri()
		if uri != self._uri or self._changed:
			self._uri = uri
			self._changed = False
			self.update()
	
	def findString(self, string):
		""" finds a string in the buffer and selects it """
		iter = self.textBuffer.get_start_iter()
		result = iter.forward_search(string, False)
		if result:
			self.textBuffer.select_range(result[0], result[1])
			self._textView.scroll_to_iter(result[0], .25)
	
	def getString(self):
		""" returns the text in the buffer """
		return self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter())
	
	def getFilename(self):
		""" returns the document's filename extracted from its URI """
		import gnomevfs
		uri = self.textBuffer.get_uri()
		if uri and gnomevfs.get_uri_scheme(uri) == "file":
			return gnomevfs.get_local_path_from_uri(uri)
		return None
	
	def commentSelection(self, uncomment=False):
		"""Comments or uncomments the current selection"""
		bounds = self.textBuffer.get_selection_bounds()
		if bounds:
			## TODO: unbind signals!
			
			itStart, itEnd = bounds
			end = self.textBuffer.create_mark(None, itEnd, True)
			
			if itStart.get_line_offset() > 0:
				itStart.set_line_index(0)
			
			maxPre = 0
			if uncomment:
				i = 0
				for line in self.textBuffer.get_text(itStart, itEnd, False).splitlines():
					j = 0
					for char in line:
						if char in ["%", " ", "\t"]:
							j += 1
						else:
							break
					if i > 0:
						if j < len(line) and line[j] not in ["%", " ", "\t"]:
							if maxPre > j:
								maxPre = j
					else:
						maxPre = j
					i += 1
				
			while itStart.compare(itEnd) < 0:
				cur = self.textBuffer.create_mark(None, itStart, True)
				if uncomment:
					if maxPre:
						itEnd = itStart.copy()
						itEnd.forward_chars(maxPre)
						self.textBuffer.delete(itStart, itEnd)
						
				else:
					self.textBuffer.insert(itStart, "%% ", 3)
					
				itStart = self.textBuffer.get_iter_at_mark(cur)
				itEnd = self.textBuffer.get_iter_at_mark(end)
				self.textBuffer.delete_mark(cur)
				itStart.forward_line()
			
			self.textBuffer.delete_mark(end)
		else:
			## no selection
			it = self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert())
			it.set_line_offset(0)
			if uncomment:
				itEnd = self.textBuffer.get_end_iter()
				string = self.textBuffer.get_text(it, itEnd, False)
				
				i = 0
				for char in string:
					if char in ["%", " ", "\t"]:
						i += 1
					else:
						break
				
				if i > 0:
					itEnd = it.copy()
					itEnd.forward_chars(i)
					self.textBuffer.delete(it, itEnd)
			else:
				self.textBuffer.insert(it, "%% ", 3)
	
	def update(self, option=const.PACKAGES|const.NEWCOMMANDS|const.STRUCTURE|const.LABELS):
		if not self.textBuffer.get_end_iter().get_offset():
			return
		if not self.isLatex():
			return
		
		string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter(), False).decode("utf-8")
		string = self.reComment.sub("", string)
		
		updateCompletion = False
		
		if option & const.STRUCTURE:
			## update the document structure
			# TODO: if we are only a fragment, then embed our structure into the one 
			# of the master document
			if not self.reStructure:
				self.reStructure = re.compile(r"\\(part|chapter|section|subsection|subsubsection|paragraph|subparagraph|include|includegraphics)\*?(\[.*?\])?(\{.*\})")
			
			self.modelStructure.clear()
		
			lastAbsLevel = 0			# last absolute level of element
			treeOffset = 0				# current offset in the tree
			parentNodes = [None]
		
			for match in self.reStructure.finditer(string):
				# match brackets
				argument = match.group(3)
				depth = 0
				i = 1
				while i < len(argument):
					if argument[i] == "{":
						depth += 1
					elif argument[i] == "}":
						depth -= 1
					i += 1
					if depth == -1:
						break
				argument = argument[1:i-1].replace("&", "&amp;")
			
				if match.group(1) == "include" or match.group(1) == "includegraphics":
					node = self.modelStructure.append(parentNodes[treeOffset], ["<span foreground=\"darkgrey\">%s</span>" % argument, match.group(0)])
				else:
					absLevel = LEVELS.index(match.group(1))		# absolute level of element
					delta = absLevel - lastAbsLevel				# delta of levels
					lastAbsLevel = absLevel
			
					if delta >= 1:
						treeOffset += 1
					else:
						treeOffset += delta
						if treeOffset < 1:
							treeOffset = 1
				
					node = self.modelStructure.append(parentNodes[treeOffset - 1], [argument, match.group(0)])
					try:
						parentNodes[treeOffset] = node
					except IndexError:
						parentNodes.append(node)
		
			self.emit("structure-updated")
		
		if option & const.PACKAGES:
			## update list of included packages
			# \usepackage commands may only be used in the preamble
			# so if we are a fragment, only the master's packages have to be parsed
			self.modelPackages.clear()
			if self._isFragment():
				masterFilename = self._getMasterFilename()
				if masterFilename:
					masterString = open(masterFilename).read().decode("utf-8")
					for match in self.rePackage.finditer(masterString):
						self.modelPackages.append([match.group(3), match.group(2)])
			else:
				for match in self.rePackage.finditer(string):
					self.modelPackages.append([match.group(3), match.group(2)])
			
			updateCompletion = True
		
		if option & const.NEWCOMMANDS:
			## update newcommands
			if not self.reNewcommand:
				self.reNewcommand = re.compile(r"\\newcommand\{(.*?)\}(\[[1-9]?\])?\{(.*)\}")
			
			self.modelNewcommands.clear()
			# \newcommandS may also be defined in the master document
			# so this must be parsed, too
			if self._isFragment():
				masterFilename = self._getMasterFilename()
				if masterFilename:
					combinedString = string + open(masterFilename).read().decode("utf-8")
					for match in self.reNewcommand.finditer(combinedString):
						self.modelNewcommands.append([match.group(1), match.group(3), "<b>%s</b>" % match.group(1), match.group(0)])
			else:
				for match in self.reNewcommand.finditer(string):
					self.modelNewcommands.append([match.group(1), match.group(3), "<b>%s</b>" % match.group(1), match.group(0)])
					
			updateCompletion = True
		
		if option & const.LABELS:
			## update labels and bibitems
			if not self.reLabelOrBibitem:
				# TODO: match the complete \bibitem to be able to show it as a hint
				self.reLabelOrBibitem = re.compile(r"\\(label|bibitem)(\[.*?\])?\{(.*?)\}")
			
			self.modelLabelsAndBibs.clear()
			# \labelS may also be defined in the master document
			# so this must be parsed, too
			if self._isFragment():
				masterFilename = self._getMasterFilename()
				if masterFilename:
					combinedString = string + open(masterFilename).read().decode("utf-8")
					for match in self.reLabelOrBibitem.finditer(combinedString):
						if match.group(1) == "label":
							self.modelLabelsAndBibs.append([match.group(3), const.LABEL, match.group(0)])
						else:
							self.modelLabelsAndBibs.append([match.group(3), const.BIBITEM, match.group(0)])
			else:
				for match in self.reLabelOrBibitem.finditer(string):
					if match.group(1) == "label":
						self.modelLabelsAndBibs.append([match.group(3), const.LABEL, match.group(0)])
					else:
						self.modelLabelsAndBibs.append([match.group(3), const.BIBITEM, match.group(0)])
		
			updateCompletion = True
		
		if updateCompletion:
			# update the CompletionStore
			self.completionProvider.update()
	
	def isLatex(self):
		"""
		Returns True if the document appears to contain LaTeX
		"""
		language = self.textBuffer.get_language()
		if language and language.get_id() == "LaTeX":
			return True
			
		uri = self.textBuffer.get_uri()
		if uri and uri.endswith(".tex"):
			return True
			
		return False
		
	def getLine(self):
		""" returns the current line number """
		return self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert()).get_line()
	
	def getIndentString(self, lineNumber=-1):
		"""
		Returns the indentation of the current line or a given line number
		"""
		if not self.reIndent:
			self.reIndent = re.compile("[ \t]+")
		
		if lineNumber > -1:
			startIter = self.textBuffer.get_iter_at_line(lineNumber)
		else:
			startIter = self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert())
			startIter.set_line_offset(0)
		endIter = startIter.copy()
		endIter.forward_to_line_end()
		string = self.textBuffer.get_text(startIter, endIter)
		match = self.reIndent.match(string)
		if match:
			return match.group()
		else:
			return ""
	
	def scrollToLine(self, line):
		"""
		Scrolls the document to a given line
		"""
		self._textView.scroll_to_iter(self.textBuffer.get_iter_at_line(line), .25)
		
	def insert(self, arg, line_number=-1):
		"""
		Inserts text into the document
			@param arg: the text to be inserted.
				If this is a list and there's _no_ selection, then its items are 
				inserted successively and the cursor is placed after the first 
				one.
				If it's a list and there _is_ a selection, then its first item 
				is inserted at the selection's start and the following ones are
				inserted at its end successively.
			@param line_number: if > -1, the text is inserted after this line
			@param restoreIndent: if True, the indentation of the current line
				is restored and continued in the following lines
		"""
		if line_number > -1:
			# get indent at given line and prepend it
			indent = self.getIndentString(line_number)
			arg = "%s%s" % (indent, "".join(arg))
			# TODO: follow indent on multiline inserts
			self.textBuffer.insert(self.textBuffer.get_iter_at_line(line_number), arg)
		else:
			indent = self.getIndentString()
			if type(arg) is str:		# one single string
				self.textBuffer.insert_at_cursor(arg.replace("\n", "\n" + indent))
			elif type(arg) is tuple:	# multiple strings
				bounds = self.textBuffer.get_selection_bounds()
				if len(bounds):
					# store the selection marks
					l_mark = self.textBuffer.create_mark(None, bounds[0], False)
					r_mark = self.textBuffer.create_mark(None, bounds[1], True)
					# insert first piece at the beginning...
					self.textBuffer.insert(bounds[0], arg[0])
					# ...and the other ones at the end
					self.textBuffer.insert(self.textBuffer.get_iter_at_mark(r_mark), "".join(arg[1:]))
					# restore selection
					self.textBuffer.select_range(self.textBuffer.get_iter_at_mark(l_mark), self.textBuffer.get_iter_at_mark(r_mark))
					# delete marks
					self.textBuffer.delete_mark(l_mark)
					self.textBuffer.delete_mark(r_mark)
				else:
					# insert first item
					head = arg[0].replace("\n", "\n" + indent)
					self.textBuffer.insert_at_cursor(head)
					# copy the insert mark
					mark = self.textBuffer.create_mark(None, self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert()), True)
					# insert the remaining pieces
					tail = "".join(arg[1:]).replace("\n", "\n" + indent)
					self.textBuffer.insert_at_cursor(tail)
					# place cursor after the first one
					self.textBuffer.place_cursor(self.textBuffer.get_iter_at_mark(mark))
					# delete mark
					self.textBuffer.delete_mark(mark)
		self.__recognizeEnvironment()
		self._textView.grab_focus()
	
	def enableHighlight(self):
		""" Sets the document's source language to LaTeX and enables highlighting """
		import gtksourceview
		
		manager = gtksourceview.SourceLanguagesManager()
		language = manager.get_language_from_mime_type("text/x-tex")
		self.textBuffer.set_language(language)
		self.textBuffer.set_highlight(True)
	
	def requirePackages(self, packages):
		if not len(packages):
			return
		
		self.update(const.PACKAGES)
		string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter())
		
		## filter packages not yet included and bundle them to one string
		foundPackages = [row[0] for row in self.modelPackages]
		filteredPackages = [p for p in packages if not p in foundPackages]
		if not len(filteredPackages):
			return
		bundle = "".join(["\\usepackage{%s}\n" % p for p in filteredPackages])
		
		## try to insert before other packages
		match = self.rePackage.search(string)
		if match:
			lineNumber = self.textBuffer.get_iter_at_offset(match.start()).get_line()
		else:
			## try to insert after the document start
			if not self.reDocument:
				self.reDocument = re.compile(r"\\documentclass(\[[A-Za-z0-9,]*\])*\{[A-Za-z0-9,]*\}")
			match = self.reDocument.search(string)
			if match:
				lineNumber = self.textBuffer.get_iter_at_offset(match.start()).get_line() + 1
			else:
				## just insert it at the start
				lineNumber = 0
		# insert
		self.insert(bundle, lineNumber)
		# update packages
		self.update(const.PACKAGES)
		# popup message
		if not self.messageDialog:
			self.messageDialog = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, None)
		if len(filteredPackages) > 1:
			self.messageDialog.set_markup("The inserted source required the additional LaTeX packages <b>%s</b>. They have been added to your document." % "</b>, <b>".join(filteredPackages))
		else:
			self.messageDialog.set_markup("The inserted source required the additional LaTeX package <b>%s</b>. It has been added to your document." % filteredPackages[0])
		self.messageDialog.run()
		self.messageDialog.hide()
	
	def __getRefCount(self, definition):
		""" Get the number of references used in a newcommand definition """
		if not self.reArgRef:
			self.reArgRef = re.compile(r"#[1-9]")
		try:
			return max([match.group()[1] for match in self.reArgRef.finditer(definition)])
		except ValueError:
			return 0
	
	def insertNewcommand(self, name, definition):
		self.update(const.NEWCOMMANDS)
		
		## check if it already exists
		if name in [row[0] for row in self.modelNewcommands]:
			message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, None)
			message.set_markup("The document already contains a new command named <b>%s</b>." % name)
			message.run()
			message.destroy()
			return
		
		string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter(), False).decode("utf-8")
		lineNumber = 0
		
		## try to insert before other newcommands
		if not self.reNewcommand:
			self.reNewcommand = re.compile(r"\\newcommand\{(.*?)\}(\[[1-9]?\])?\{(.*)\}")
		match = self.reNewcommand.search(string)
		if match:
			lineNumber = self.textBuffer.get_iter_at_offset(match.start()).get_line()
		else:
			## try to insert after the document start
			if not self.reDocument:
				self.reDocument = re.compile(r"\\documentclass(\[[A-Za-z0-9,]*\])*\{[A-Za-z0-9,]*\}")
			match = self.reDocument.search(string)
			if match:
				lineNumber = self.textBuffer.get_iter_at_offset(match.start()).get_line() + 1
			else:
				## just insert it at the start
				lineNumber = 0
		refStr = ""
		refCount = self.__getRefCount(definition)
		if refCount:
			refStr = "[%s]" % refCount
		self.insert("\\newcommand{%s}%s{%s}\n" % (name, refStr, definition), lineNumber)
		self.update(const.NEWCOMMANDS)
	
	def deleteNewcommand(self, name):
		self.update(const.NEWCOMMANDS)
		
		## load string
		string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter(), False).decode("utf-8")
		if self._isFragment():
			masterFilename = self._getMasterFilename()
			if masterFilename:
				string += open(masterFilename).read().decode("utf-8")
		string = self.reComment.sub("", string)
		
		## count references
		c = string.count(name) - 1	# the first one is the definition itself
		if c:
			message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, None)
			message.set_markup("The new command <b>%s</b> is referenced <b>%s</b> times in this document or its master.\nAre you sure you want to delete it?" % (name, c))
			if message.run() == gtk.RESPONSE_NO:
				message.destroy()
				return
			message.destroy()
		
		## delete the definition
		pattern = re.compile(r"\\newcommand\{%s\}(\[[1-9]?\])?\{(.*)\}" % re.escape(name))
		match = pattern.search(string)
		if match:
			self.textBuffer.select_range(self.textBuffer.get_iter_at_offset(match.start()), self.textBuffer.get_iter_at_offset(match.end()))
			self.textBuffer.delete_selection(True, True)
		
		self.update(const.NEWCOMMANDS)
	
	def changeNewcommand(self, originalName, newName, definition):
		if newName != originalName:
			## change all occurences of the newcommand
			## apply regex to the string and then set_text() it again
			print "Changing the name of a newcommand is not yet implemented."
		else:
			## only change the newcommand's definition and number of references
			pattern = re.compile(r"\\newcommand\{%s\}(\[[1-9]?\])?\{(.*)\}" % re.escape(originalName))
			string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_end_iter(), False).decode("utf-8")
			match = pattern.search(string)
			if match:
				self.textBuffer.select_range(self.textBuffer.get_iter_at_offset(match.start()), self.textBuffer.get_iter_at_offset(match.end()))
				self.textBuffer.delete_selection(True, True)
				## build newcommand
				refStr = ""
				refCount = self.__getRefCount(definition)
				if refCount:
					refStr = "[%s]" % refCount
				command = "\\newcommand{%s}%s{%s}" % (originalName, refStr, definition)
				self.textBuffer.insert_at_cursor(command)
		self.update(const.NEWCOMMANDS)
	
	def __recognizeEnvironment(self):
		""" Recognize the environment the cursor is in """
		# FIXME: only parse when the buffer has changed or the cursor has moved
		# (don't parse on Ctrl or Alt)
		
		if self.isLatex():
			string = self.textBuffer.get_text(self.textBuffer.get_start_iter(), self.textBuffer.get_iter_at_mark(self.textBuffer.get_insert()), False)
			## remove comments
			string = self.reComment.sub("", string)
			
			## math
			counter = 0
			for item in [r"$", r"{math", r"{displaymath", r"{eqnarray", r"\[", r"\]", r"{align", r"{equation"]:
				counter += string.count(item)
			
			if counter % 2:
				environment = const.MATH
			else:
				environment = const.GENERAL
			
			## tabular
			c = 0
			for item in ["{tabular", "{matrix", "{pmatrix", "{bmatrix", "{Bmatrix", "{vmatrix", "{Vmatrix"]:
				c += string.count(item)
			if c % 2:
				environment = environment|const.TABULAR
			
			if environment != self.environment:
				self.environment = environment
				self.emit("environment-changed")
		else:
			environment = const.UNKNOWN
			if environment != self.environment:
				self.environment = environment
				self.emit("environment-changed")
				
	def __onBufferChanged(self, textBuffer):
		self._changed = True

# register GObject
gobject.type_register(DocumentWrapper)
