# -*- coding: utf-8 -*-

# __init__.py
# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2006 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

import gedit, gtk, os.path
from config import Preferences
from DocumentWrapper import DocumentWrapper
from StructurePanel import StructurePanel
from SymbolsPanel import SymbolsPanel
from macros import MacrosPanel
from completion import CompletionPopup
import const

DBUS_BUS_NAME = 'org.gedit.LaTeXPlugin'
DBUS_OBJECT_PATH = '/org/gedit/DBUSInterface'

CUSTOM_ICONS = {
		"/LaTeXToolbar/QuickstartTool" : "quickstart.png",
		
		"/LaTeXToolbar/BoldTool" : "bf.png",
		"/LaTeXToolbar/ItalicTool" : "it.png",
		"/LaTeXToolbar/UnderlineTool" : "underline.png",
		"/LaTeXToolbar/SmallCapsTool" : "sc.png",
		
		"/LaTeXToolbar/CenterTool" : "center.png",
		
		"/LaTeXToolbar/ItemizeTool" : "itemize.png",
		"/LaTeXToolbar/EnumerateTool" : "enumerate.png",
		"/LaTeXToolbar/DescriptionTool" : "description.png",
		"/LaTeXToolbar/ItemTool" : "item.png",
		
		"/LaTeXToolbar/SubscriptTool" : "sub.png",
		"/LaTeXToolbar/SuperscriptTool" : "super.png",
		"/LaTeXToolbar/FractionTool" : "frac.png",
		"/LaTeXToolbar/GraphicsTool" : "graphics.png",
		"/LaTeXToolbar/TableTool" : "table.png",
		
		"/TypefacePopup/TypefaceRmMenu" : "rm.png",
		"/TypefacePopup/TypefaceSfMenu" : "sf.png",
		"/TypefacePopup/TypefaceTtMenu" : "tt.png",
		"/TypefacePopup/TypefaceBbMenu" : "bb.png",
		"/TypefacePopup/TypefaceCalMenu" : "cal.png",
		"/TypefacePopup/TypefaceFrakMenu" : "frak.png",
		
		"/StructurePopup/StructurePartMenu" : "part.png",
		"/StructurePopup/StructureChapterMenu" : "chapter.png",
		"/StructurePopup/StructureSectionMenu" : "section.png",
		"/StructurePopup/StructureSubsectionMenu" : "subsection.png",
		"/StructurePopup/StructureSubsubsectionMenu" : "subsubsection.png",
		"/StructurePopup/StructureParagraphMenu" : "paragraph.png",
		"/StructurePopup/StructureSubparagraphMenu" : "paragraph.png",
		
		"/MathMenu/DisplayMathMenu" : "displaymath.png",
		"/MathMenu/EquationMenu" : "equation.png",
		"/MathMenu/UnnumberedEqnArrayMenu" : "uneqnarray.png",
		"/MathMenu/NumberedEqnArrayMenu" : "eqnarray.png",
		
		"/AccentPopup/AccentDotMenu" : "dot.png",
		"/AccentPopup/AccentDdotMenu" : "ddot.png",
		"/AccentPopup/AccentBarMenu" : "bar.png",
		"/AccentPopup/AccentTildeMenu" : "tilde.png",
		"/AccentPopup/AccentVecMenu" : "vec.png",
		"/AccentPopup/AccentHatMenu" : "hat.png",
		"/AccentPopup/AccentAcuteMenu" : "acute.png",
		"/AccentPopup/AccentGraveMenu" : "grave.png",
		"/AccentPopup/AccentBreveMenu" : "breve.png",
		"/AccentPopup/AccentCheckMenu" : "check.png",
		
		"/CompilePopup/CompilePDFMenu" : "compile-pdf.png",
		"/CompilePopup/CompileDVIMenu" : "compile-dvi.png",
		"/CompilePopup/CompilePSMenu" : "compile-ps.png",
		"/CompilePopup/CompileImageMenu" : "compile-image.png"
}

class C:
	"""
	Encapsulates a LaTeX macro or command
	"""
	def __init__(self, components, package=None):
		self._components = components
		self._package = package
	def __getitem__(self, index):
		return self._components[index]
	def __len__(self):
		return len(self._components)
	def get_package(self):
		return self._package

ACTIONS_TO_MACROS = {
		"Bold" : (C((r"\textbf{", "}")), C((r"\mathbf{", "}"))),
		"Italic" : (C((r"\textit{", "}")), C((r"\mathit{", "}"))),
		"Underline" : (C((r"\underline{", "}")), C((r"\underline{", "}"))),
		"SmallCaps" : (C((r"\textsc{", "}")), None),
		
		"Center" : (C(("\\begin{center}\n\t", "\n\\end{center}")), None),
		
		"Itemize" : (C(("\\begin{itemize}\n\t\\item", "\n\\end{itemize}")), None),
		"Enumerate" : (C(("\\begin{enumerate}\n\t\\item", "\n\\end{enumerate}")), None),
		"Description" : (C(("\\begin{description}\n\t\\item[", "]\n\\end{description}")), None),
		"Item" : (C(("\\item ")), None),
		
		"DisplayMath" : (C(("\[ ", " \]")), None),
		"Equation" : (C(("\\begin{equation}\n\t", "\n\\end{equation}")), None),
		"UnnumberedEqnArray" : (C(("\\begin{align*}\n\t", "\n\\end{align*}"), "amsmath"), None),
		"NumberedEqnArray" : (C(("\\begin{align}\n\t", "\n\\end{align}"), "amsmath"), None),
		
		"TypefaceRm" : (C(("\\textrm{", "}")), C(("\\mathrm{", "}"))),
		"TypefaceSf" : (C(("\\textsf{", "}")), C(("\\mathsf{", "}"))),
		"TypefaceTt" : (C(("\\texttt{", "}")), C(("\\mathtt{", "}"))),
		"TypefaceBb" : (None, C(("\\mathbb{", "}"), "amssymb")),
		"TypefaceCal" : (None, C(("\\mathcal{", "}"))),
		"TypefaceFrak" : (None, C(("\\mathfrak{", "}"), "amssymb")),
		
		"StructurePart" : (C(("\\part{", "}")), None),
		"StructureChapter" : (C(("\\chapter{", "}")), None),
		"StructureSection" : (C(("\\section{", "}")), None),
		"StructureSubsection" : (C(("\\subsection{", "}")), None),
		"StructureSubsubsection" : (C(("\\subsubsection{", "}")), None),
		"StructureParagraph" : (C(("\\paragraph{", "}")), None),
		"StructureSubparagraph" : (C(("\\subparagraph{", "}")), None),
		
		"Subscript" : (None, C((r"_{", "}"))),
		"Superscript" : (None, C((r"^{", "}"))),
		"Fraction" : (None, C((r"\frac{", "}{}"))),
		
		"AccentDot" : (C(("\\.{", "}")), C(("\\dot{", "}"))),
		"AccentDdot" : (C(('\\"{', "}")), C(("\\ddot{", "}"))),
		"AccentBar" : (None, C(("\\bar{", "}"))),
		"AccentTilde" : (C(("\\~{", "}")), C(("\\tilde{", "}"))),
		"AccentVec" : (None, C(("\\vec{", "}"))),
		"AccentHat" : (C(("\\^{", "}")), C(("\\hat{", "}"))),
		"AccentAcute" : (C(("\\'{", "}")), C(("\\acute{", "}"))),
		"AccentGrave" : (C(("\\`{", "}")), C(("\\grave{", "}"))),
		"AccentBreve" : (None, C(("\\breve{", "}"))),
		"AccentCheck" : (None, C(("\\check{", "}"))),
}

# widgets that are only defined in special LaTeX environments
SMART_WIDGETS = {
		"/LaTeXToolbar/SmallCapsTool" : const.GENERAL,
		"/LaTeXToolbar/ItemizeTool" : const.GENERAL,
		"/LaTeXToolbar/EnumerateTool" : const.GENERAL,
		"/LaTeXToolbar/DescriptionTool" : const.GENERAL,
		"/LaTeXToolbar/ItemTool" : const.GENERAL,
		"/LaTeXToolbar/SubscriptTool" : const.MATH,
		"/LaTeXToolbar/SuperscriptTool" : const.MATH,
		"/LaTeXToolbar/FractionTool" : const.MATH,
		"/AccentPopup/AccentBarMenu" : const.MATH,
		"/AccentPopup/AccentVecMenu" : const.MATH,
		"/AccentPopup/AccentBreveMenu" : const.MATH,
		"/AccentPopup/AccentCheckMenu" : const.MATH,
		"/TypefacePopup/TypefaceBbMenu" : const.MATH,
		"/TypefacePopup/TypefaceCalMenu" : const.MATH,
		"/TypefacePopup/TypefaceFrakMenu" : const.MATH,
		"/LaTeXToolbar/GraphicsTool" : const.GENERAL
}

class LaTeXPluginInstance:
	wrappers = {}			# maps the GeditViews to DocumentWrappers
	lastWrapper = None		# the last DocumentWrapper activated
	wrapperConnections = {}	# maps signals to handler ids for reconnecting
	
	quickstartDialog, graphicsDialog, tableDialog, backend, backendDialog, renderImageDialog = None, None, None, None, None, None
	
	def __init__(self, plugin, window):
		self.plugin, self.window = plugin, window
		
		self.preferences = Preferences()
		self.uiManager = window.get_ui_manager()
		self.saveAction = self.uiManager.get_action("/MenuBar/FileMenu/FileSaveMenu")
		self.structurePanel = StructurePanel(self)
		self.structurePanel.set_position(self.preferences.setdefault("/structure-paned-position", -1))
		self.symbolMapPanel = SymbolsPanel(self)
		self.macrosPanel = MacrosPanel(self)
		self.macrosPanel.set_position(self.preferences.setdefault("/macros-paned-position", -1))
		
		self.primaryClipboard = gtk.clipboard_get("PRIMARY")
		
		self.window.get_side_panel().add_item(self.structurePanel, _("Document Structure"), gtk.image_new_from_stock(gtk.STOCK_INDEX, gtk.ICON_SIZE_MENU))
		self.window.get_side_panel().add_item(self.symbolMapPanel, _("Common Symbols"), gtk.image_new_from_stock(gtk.STOCK_SELECT_FONT, gtk.ICON_SIZE_MENU))
		self.window.get_side_panel().add_item(self.macrosPanel, _("Macros and New Commands"), gtk.image_new_from_stock(gtk.STOCK_EXECUTE, gtk.ICON_SIZE_MENU))
		
		self.builtinCompletionTargets = open(const.PATH + "/completion/builtin").read().splitlines()
		self.completionPopup = CompletionPopup()
		
		self.__insertUI()
		
		window.connect("destroy", lambda a: self.__cleanup())
		
		self.dviSearchEnabled = self.preferences.setdefault("/dvi-search-enabled", False)
		if self.dviSearchEnabled:
			# load DBUS modules and create a service
			try:
				import dbus.service
				if getattr(dbus, 'version', (0,0,0)) >= (0,41,0):
					import dbus.glib
				dbusImported = True
			except ImportError:
				print "Failed to import dbus"
				dbusImported = False
			
			if dbusImported:
				class DBUSInterface(dbus.service.Object):
					"""
					The DBUSInterface object is listening for DVI inverse search commands and
					scrolls the document to the appropriate line
					"""
					def __init__(self, instance, bus_name, object_path=DBUS_OBJECT_PATH):
						dbus.service.Object.__init__(self, bus_name, object_path)
						self._instance = instance

					# This is for dbus inverse search
					@dbus.service.method('org.gedit.LaTeXPluginIFace')
					def inverse(self, filename, line):
						print "Got inverse DVI search: " + filename + " " + str(line)
						# TODO: switch to the appropriate tab if the file is not active
						document = self._instance.getDocument()
						# FIXME: document.tempFilename is not used anymore
						# Backend now handles temp files on its own
						if os.path.samefile(document.getFilename(), filename):
							document.scrollToLine(line - 1)		# tex counts from 1
						else:	
							print "Error: file not loaded or active"

					# adding this makes things work---don't know why!
					@dbus.service.method('org.gedit.LaTeXPluginIFace')
					def dummy(self):
						pass
			
				# DBUS interface
				bus_name = dbus.service.BusName(DBUS_BUS_NAME, bus=dbus.SessionBus())
				self._dbus = DBUSInterface(self, bus_name)
	
	def getDocument(self):
		""" returns the DocumentWrapper of currently selected GeditView or None if
			there are no tabs """
		activeView = self.window.get_active_view()
		if activeView:
			try:
				return self.wrappers[activeView]
			except KeyError:
				return None
		return None
	
	def __insertUI(self):
		""" merges the plugin's UI """
		self.actionGroup = gtk.ActionGroup("LaTeXPluginActions")
		self.actionGroup.add_actions([
				("Comment", None, _("Comment"), "<Ctrl><Alt>C", None, lambda a: self.getDocument().commentSelection()),
				("Uncomment", None, _("Uncomment"), "<Ctrl><Alt>U", None, lambda a: self.getDocument().commentSelection(True)),
				
				("RefreshStructure", gtk.STOCK_REFRESH, _("Refresh Document Structure"), "F5", _("Parse the document again and refresh the structure view"), lambda a: self.getDocument().update()),
				("FindInDVI", gtk.STOCK_FIND, _("Find selection in DVI"), "<Shift><Ctrl>F", _("Finds the current position in the DVI document"), lambda a: self.__performDVISearch()),
				("SearchPrimary", gtk.STOCK_FIND, _("Find primary selection"), "<Shift><Ctrl>Q", _("Finds the primary selection in the current document"), lambda a: self.__performPrimarySearch()),
				("Compile", gtk.STOCK_CONVERT, _("Compile LaTeX"), None, None, None),
				("CompilePDF", None, "PDF", "F6", _("Compiles the document to PDF"), lambda a: self.__compileDocument(const.PDF)),
				("CompileDVI", None, "DVI", "<Ctrl>F6", _("Compiles the document to DVI"), lambda a: self.__compileDocument(const.DVI)),
				("CompilePS", None, "PostScript", "<Shift><Ctrl>F6", _("Compiles the document to PostScript"), lambda a: self.__compileDocument(const.PS)),
				("CompileImage", None, _("Image") + "...", None, _("Renders the document to an image"), lambda a: self.__renderToImage()),
				
				("Quickstart", gtk.STOCK_MISSING_IMAGE, _("Quickstart"), None, _("Create a new LaTeX document"), lambda a: self.__runQuickstartDialog()),
				("Bold", gtk.STOCK_MISSING_IMAGE, _("Bold"), None, _("Bold"), self.__insertMacro),
				("Italic", gtk.STOCK_MISSING_IMAGE, _("Italic"), None, _("Italic"), self.__insertMacro),
				("Underline", gtk.STOCK_MISSING_IMAGE, _("Underline"), None, _("Underline"), self.__insertMacro),
				("SmallCaps", gtk.STOCK_MISSING_IMAGE, _("Small Caps"), None, _("Small Capitals"), self.__insertMacro),
				
				("Center", gtk.STOCK_JUSTIFY_CENTER, _("Center"), None, _("Center"), self.__insertMacro),
				
				("Itemize", gtk.STOCK_MISSING_IMAGE, _("List"), None, _("Insert a list"), self.__insertMacro),
				("Enumerate", gtk.STOCK_MISSING_IMAGE, _("Enumeration"), None, _("Insert an enumeration"), self.__insertMacro),
				("Description", gtk.STOCK_MISSING_IMAGE, _("Descriptions"), None, _("Insert a description list"), self.__insertMacro),
				("Item", gtk.STOCK_MISSING_IMAGE, _("Item"), None, _("Insert an item into a list or enumeration"), self.__insertMacro),
				
				("Graphics", gtk.STOCK_MISSING_IMAGE, _("Graphics"), None, _("Include a graphics"), lambda a: self.__insertGraphics()),
				("Table", gtk.STOCK_MISSING_IMAGE, _("Table"), None, _("Insert a table or matrix"), lambda a: self.__insertTable()),
				("Subscript", gtk.STOCK_MISSING_IMAGE, _("Subscript"), None, _("Subscript"), self.__insertMacro),
				("Superscript", gtk.STOCK_MISSING_IMAGE, _("Superscript"), None, _("Superscript"), self.__insertMacro),
				("Fraction", gtk.STOCK_MISSING_IMAGE, _("Fraction"), None, _("Fraction"), self.__insertMacro),
				
				("TypefaceRm", None, _("Roman"), None, None, self.__insertMacro),
				("TypefaceSf", None, _("Sans serif"), None, None, self.__insertMacro),
				("TypefaceTt", None, _("Monospace"), None, None, self.__insertMacro),
				("TypefaceBb", None, _("Blackboard Bold (Set)"), None, None, self.__insertMacro),
				("TypefaceCal", None, _("Caligraphy"), None, None, self.__insertMacro),
				("TypefaceFrak", None, _("Fraktur"), None, None, self.__insertMacro),
				
				("StructurePart", None, _("Part"), None, None, self.__insertMacro),
				("StructureChapter", None, _("Chapter"), None, None, self.__insertMacro),
				("StructureSection", None, _("Section"), None, None, self.__insertMacro),
				("StructureSubsection", None, _("Subsection"), None, None, self.__insertMacro),
				("StructureSubsubsection", None, _("Sub-subsection"), None, None, self.__insertMacro),
				("StructureParagraph", None, _("Paragraph"), None, None, self.__insertMacro),
				("StructureSubparagraph", None, _("Subparagraph"), None, None, self.__insertMacro),
				
				("DisplayMath", None, _("Centered Formula"), None, None, self.__insertMacro),
				("Equation", None, _("Numbered Equation"), None, None, self.__insertMacro),
				("UnnumberedEqnArray", None, _("Array of Equations"), None, None, self.__insertMacro),
				("NumberedEqnArray", None, _("Array of Numbered Equations"), None, None, self.__insertMacro),
				
				("AccentDot", None, _("Dot"), None, None, self.__insertMacro),
				("AccentDdot", None, _("Double dot"), None, None, self.__insertMacro),
				("AccentBar", None, _("Bar"), None, None, self.__insertMacro),
				("AccentTilde", None, _("Tilde"), None, None, self.__insertMacro),
				("AccentVec", None, _("Vector"), None, None, self.__insertMacro),
				("AccentHat", None, _("Hat"), None, None, self.__insertMacro),
				("AccentAcute", None, _("Acute"), None, None, self.__insertMacro),
				("AccentGrave", None, _("Grave"), None, None, self.__insertMacro),
				("AccentBreve", None, _("Breve"), None, None, self.__insertMacro),
				("AccentCheck", None, _("Hacek"), None, None, self.__insertMacro),
				
				("NewcommandInsert", gtk.STOCK_GO_FORWARD, _("Paste Command"), None, None, self.macrosPanel.onNewcommandContextInsert),
				("NewcommandJump", gtk.STOCK_JUMP_TO, _("Jump to Definition"), None, None, self.macrosPanel.onNewcommandContextJump),
				("NewcommandEdit", gtk.STOCK_EDIT, _("Edit..."), None, None, self.macrosPanel.onNewcommandContextEdit),
				("NewcommandDelete", gtk.STOCK_DELETE, _("Delete"), None, None, self.macrosPanel.onNewcommandContextDelete),
				
				("MacroInsert", gtk.STOCK_GO_FORWARD, _("Paste Definition"), None, None, self.macrosPanel.onMacroContextInsert),
				("MacroEdit", gtk.STOCK_EDIT, _("Edit..."), None, None, self.macrosPanel.onMacroContextEdit),
				("MacroDelete", gtk.STOCK_DELETE, _("Delete"), None, None, self.macrosPanel.onMacroContextDelete)
		])
		self.uiManager.insert_action_group(self.actionGroup, -1)
		self.uiID = self.uiManager.add_ui_from_file(const.PATH + "/ui.xml")
		
		self.uiManager.get_widget("/MenuBar/SearchMenu/FindInDVIMenu").set_sensitive(False)
		
		# insert the toolbar
		box = self.window.get_children()[0]
		self.toolbar = self.uiManager.get_widget("/LaTeXToolbar")
		box.pack_start(self.toolbar, False)
		box.reorder_child(self.toolbar, 2)
		self.toolbar.set_style(gtk.TOOLBAR_ICONS)
		self.toolbar.show()
		
		# insert the toolbar CheckMenuItem
		self.menuViewToolbar = gtk.CheckMenuItem(_("LaTeX toolbar"))
		self.menuViewToolbar.set_active(True)
		self.uiManager.get_widget("/MenuBar/ViewMenu").get_submenu().insert(self.menuViewToolbar, 2)
		self.menuViewToolbar.show()
		self.menuViewToolbar.connect("toggled", self.__onViewToolbarToggled)
		
		# set custom icons
		for path in CUSTOM_ICONS.keys():
			widget = self.uiManager.get_widget(path)
			image = gtk.Image()
  			image.set_from_file("%s/pixmaps/%s" % (const.PATH, CUSTOM_ICONS[path]))
  			image.show()
			if type(widget) is gtk.ImageMenuItem:
				widget.set_image(image)
			elif type(widget) is gtk.ToolButton:
				widget.set_icon_widget(image)
		
		# insert typeface SmartMenuToolButton
		self.buttonTypeface = SmartMenuToolButton(_("Typeface"), self.uiManager.get_widget("/TypefacePopup"), 2)
		self.buttonTypeface.show()
		self.toolbar.insert(self.buttonTypeface, 6)
		
		# insert structure SmartMenuToolButton
		self.buttonStructure = SmartMenuToolButton(_("Headline"), self.uiManager.get_widget("/StructurePopup"), 3)
		self.buttonStructure.show()
		self.toolbar.insert(self.buttonStructure, 15)
		
		# insert math MenuToolButton
		image = gtk.Image()
		image.set_from_file(const.PATH + "/pixmaps/math.png")
		self.buttonMath = gtk.MenuToolButton(image, _("Mathematical"))
		self.buttonMath.set_menu(self.uiManager.get_widget("/MathMenu"))
		self.buttonMath.connect("clicked", lambda a: self.getDocument().insert(("$", "$")))
		self.buttonMath.show()
		self.toolbar.insert(self.buttonMath, 17)
		
		# insert accents SmartMenuToolButton
		self.buttonAccent = SmartMenuToolButton(_("Accent"), self.uiManager.get_widget("/AccentPopup"))
		self.buttonAccent.show()
		self.toolbar.insert(self.buttonAccent, 21)
		
		# insert compile SmartMenuToolButton
		self.buttonCompile = SmartMenuToolButton(_("Compile"), self.uiManager.get_widget("/CompilePopup"))
		self.buttonCompile.show()
		self.toolbar.insert(self.buttonCompile, 26)
		
		self.menuViewToolbar.set_active(self.preferences.setdefault("/toolbar-visible", True))
	
	def __onViewToolbarToggled(self, checkmenuitem):
		"""
		Shows or hides the toolbar
		"""
		if checkmenuitem.get_active():
			self.toolbar = self.uiManager.get_widget("/LaTeXToolbar").show()
		else:
			self.toolbar = self.uiManager.get_widget("/LaTeXToolbar").hide()
	
	def __insertMacro(self, action):
		environment = self.getDocument().environment
		if environment == const.MATH:
			macro = ACTIONS_TO_MACROS[action.get_name()][1]
		else:
			macro = ACTIONS_TO_MACROS[action.get_name()][0]
		if not macro:
			return
		package = macro.get_package()
		if package:
			self.getDocument().requirePackages([package])
		self.getDocument().insert(tuple(macro))
	
	def __runQuickstartDialog(self):
		if not self.quickstartDialog:
			from QuickstartDialog import QuickstartDialog
			self.quickstartDialog = QuickstartDialog(self)
		if self.quickstartDialog.run() == gtk.RESPONSE_OK:
			document = self.getDocument()
			if len(document.getString()) or document.getFilename():
				question = gtk.MessageDialog(self.window, gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT, 
						gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, _("Should I create a new gedit document?"))
				if question.run() == gtk.RESPONSE_YES:
					self.window.create_tab(True)
					document = self.getDocument()
				question.destroy()
			document.insert(self.quickstartDialog.getSource())
			document.enableHighlight()
			document.update()
		self.quickstartDialog.hide()
	
	def __insertGraphics(self):
		if not self.graphicsDialog:
			from GraphicsDialog import GraphicsDialog
			dpi = self.preferences.setdefault("/dpi", 96)
			self.graphicsDialog = GraphicsDialog(self.window, dpi)
		if self.graphicsDialog.run() == gtk.RESPONSE_OK:
			self.getDocument().insert(self.graphicsDialog.source)
			self.getDocument().requirePackages(["graphicx"])
		self.graphicsDialog.hide()
	
	def __insertTable(self):
		if not self.tableDialog:
			from TableDialog import TableDialog
			self.tableDialog = TableDialog()
		if self.tableDialog.run(self.getDocument().environment) == gtk.RESPONSE_OK:
			self.getDocument().insert(self.tableDialog.source)
			if self.tableDialog.package:
				self.getDocument().requirePackages([self.tableDialog.package])
		self.tableDialog.hide()
	
	def __renderToImage(self):
		if not self.renderImageDialog:
			from RenderImageDialog import RenderImageDialog
			self.renderImageDialog = RenderImageDialog()
		if self.renderImageDialog.run() == gtk.RESPONSE_OK:
			if not self.backend:
				self.backend = Backend()
				self.backendDialog = BackendDialog(self.backend, self)
			self.backend.setImageParameters(self.renderImageDialog.pngMode, self.renderImageDialog.aaFactor, 
					self.renderImageDialog.resolution, self.renderImageDialog.boundingBox)
			self.__compileDocument(self.renderImageDialog.format)
		self.renderImageDialog.hide()
	
	def __compileDocument(self, format):
		# check if rubber is installed
		if os.system("rubber --version $2>/dev/null"):
			message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, None)
			message.set_markup("The LaTeX plugin requires <b>rubber</b> for automated document compiling.\nThis tool can be downloaded from <u>rubber.sourceforge.net</u>.")
			message.run()
			message.destroy()
			return
			
		# create a Backend and BackendDialog instance if neccessary
		if not self.backend:
			from backend import RubberBackend, BackendDialog
			self.backend = RubberBackend()
			self.backendDialog = BackendDialog(self.backend, self)
		
		document = self.getDocument()
		filename = document.getFilename()
		if filename:
			# save the document (if it has been changed since the last save)
			self.saveAction.activate()
			# compile it
			if not self.backend.compile(format, filename=filename):
				# compile() only returns False if the master document could not be found
				message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, None)
				message.set_markup("The current document seems to be included by another one, which could not be found though.\n\n<b>Hint 1:</b> The parent document and all included child documents must be in the same directory.\n<b>Hint 2:</b> \\include commands can not be nested.")
				message.run()
				message.destroy()
				return
		else:
			self.backend.compile(format, string=document.getString())
		
		## enable/disable DVI search
		document.dviCompiled = (format == const.DVI)
	
	def __performDVISearch(self):
		if not self.backend:
			return
		self.backend.findLine(self.getDocument().getLine() + 1)
	
	def __performPrimarySearch(self):
		""" Searches the active document for the string from the primary clipboard """
		text = self.primaryClipboard.wait_for_text()
		# Here we should be doing some cleaning-up
		# Remove hyphenation in mid-word at line end
		text = text.replace('-\n', '')
		# Replace CRs with blanks and elim needless blanks
		text = text.replace('\n', ' ').strip()
		# Double quotes
		text = text.replace('\xe2\x80\x9c', '``')
		text = text.replace('\xe2\x80\x9d', "''")
		# Ligatures in UTF-8 encoding
		# FIXME: can we use text.translate() ?
		text = text.replace('\xef\xac\x80', 'ff')
		text = text.replace('\xef\xac\x81', 'fi')
		text = text.replace('\xef\xac\x82', 'fl')
		text = text.replace('\xef\xac\x83', 'ffi')
		text = text.replace('\xef\xac\x84', 'ffl')
		self.getDocument().findString(text)
	
	def __removeUI(self):
		""" Removes the plugin's ui """
		self.window.get_side_panel().remove_item(self.symbolMapPanel)
		self.window.get_side_panel().remove_item(self.structurePanel)
		self.window.get_side_panel().remove_item(self.macrosPanel)
		self.toolbar = None
		self.menuViewToolbar.destroy()
		self.uiManager.remove_ui(self.uiID)
		self.uiManager.remove_action_group(self.actionGroup)
		self.uiManager.ensure_update()
	
	def deactivate(self):
		""" Removes the ui and destroys some objects """
		self.__removeUI()
		self.window = None
		self.plugin = None
		self.actionGroup = None
		self.uiManager = None
		
	def updateUI(self):
		""" Deactivate the ui if there is no document """
		activeView = self.window.get_active_view()
		
		self.buttonTypeface.set_sensitive(bool(activeView))
		self.buttonStructure.set_sensitive(bool(activeView))
		self.buttonMath.set_sensitive(bool(activeView))
		self.buttonAccent.set_sensitive(bool(activeView))
		self.buttonCompile.set_sensitive(bool(activeView))
		self.actionGroup.set_sensitive(bool(activeView))
		
		if not activeView:
			self.structurePanel.viewStructure.set_model(None)
			self.structurePanel.viewLabels.set_model(None)
			self.macrosPanel.viewNewcommands.set_model(None)
			return
		
		## enable/disable DVI search menu entry
		document = self.getDocument()
		if document:
			self.uiManager.get_widget("/MenuBar/SearchMenu/FindInDVIMenu").set_sensitive(self.dviSearchEnabled and document.dviCompiled)
		
		try:
			wrapper = self.wrappers[activeView]
		except KeyError:
			wrapper = DocumentWrapper(activeView, self)
			self.wrappers[activeView] = wrapper
		
		if wrapper is self.lastWrapper:
			wrapper.activate()
			self.__updateSmartWidgets()
			return
		
		# disconnect the last wrapper
		if self.lastWrapper:
			self.lastWrapper.disconnect(self.wrapperConnections["environment-changed"])
			self.lastWrapper.disconnect(self.wrapperConnections["structure-updated"])
		# connect to the signals of the wrapper
		self.wrapperConnections["environment-changed"] = wrapper.connect("environment-changed", lambda a: self.__updateSmartWidgets())
		self.wrapperConnections["structure-updated"] = wrapper.connect("structure-updated", lambda a: self.structurePanel.viewStructure.expand_all())
		# pass the structure
		self.structurePanel.viewStructure.set_model(wrapper.modelStructure)
		self.structurePanel.viewStructure.expand_all()
		# labels
		self.structurePanel.viewLabels.set_model(wrapper.modelLabelsAndBibs)
		self.macrosPanel.viewNewcommands.set_model(wrapper.modelNewcommands)
		
		self.lastWrapper = wrapper
		self.__updateSmartWidgets()
	
	def __cleanup(self):
		""" Stores preferences """
		self.preferences["/macros-paned-position"] = self.macrosPanel.get_position()
		self.preferences["/structure-paned-position"] = self.structurePanel.get_position()
		self.preferences["/toolbar-visible"] = self.menuViewToolbar.get_active()
	
	def __updateSmartWidgets(self):
		""" (de)activates widgets that are only defined in special environments """
		environment = self.getDocument().environment
		for path in SMART_WIDGETS:
			if environment == const.UNKNOWN:
				self.uiManager.get_widget(path).set_sensitive(True)
			else:
				if environment & SMART_WIDGETS[path]:
					self.uiManager.get_widget(path).set_sensitive(True)
				else:
					self.uiManager.get_widget(path).set_sensitive(False)
	
class SmartMenuToolButton(gtk.MenuToolButton):
	""" A custom MenuToolButton that switches the icon and action of its button 
		to the last selected ones """
	activeItem = None
	
	def __init__(self, label, menu, index=0):
		item = menu.get_children()[index]
		self.image = gtk.Image()
		self.image.set_from_pixbuf(item.get_image().get_pixbuf())
		self.image.show()
		self.activeItem = item
		
		gtk.MenuToolButton.__init__(self, self.image, label)
		self.button = self.get_children()[0].get_children()[0]
		
		self.__set_menu(menu)
	
	def __set_menu(self, menu):
		children = menu.get_children()
		for child in children:
			child.connect("activate", self.__onItemActivated)
		gtk.MenuToolButton.set_menu(self, menu)
		self.connect("clicked", self.__onButtonClicked)
		
	def __onItemActivated(self, item):
		""" called when a menu item is activated """
		self.image.set_from_pixbuf(item.get_image().get_pixbuf())
		self.button.set_image(self.image)
		self.activeItem = item

	def __onButtonClicked(self, button):
		""" called when the MenuToolButton is clicked """
		if self.activeItem:
			self.activeItem.emit("activate")

class LaTeXPlugin(gedit.Plugin):
	TAG = "LaTeXPluginInstance"
	
	def __init__(self):
		gedit.Plugin.__init__(self)
	
	def __getInstance(self, window):
		return window.get_data(self.TAG)
	
	def __setInstance(self, window, instance):
		window.set_data(self.TAG, instance)
	
	def activate(self, window):
		"""
		called when a new gedit window is created
		"""
		self.__setInstance(window, LaTeXPluginInstance(self, window))
	
	def deactivate(self, window):
		"""
		called when a window is destroyed
		"""
		self.__getInstance(window).deactivate()
		self.__setInstance(window, None)
	
	def update_ui(self, window):
		"""
		called when the window is updated (tab change, menu update etc.)
		"""
		self.__getInstance(window).updateUI()
	
	def create_configure_dialog(self):
		"""
		called when the plugin is to be configured (must return a gtk.Window)
		"""
		from config import ConfigDialog
		return ConfigDialog()
